// Copyright 'dart-universal_io' project authors.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

import 'dart:async';
import 'dart:convert';
import 'dart:html' as html;
import 'dart:typed_data';

import 'package:meta/meta.dart';
import 'package:typed_data/typed_buffers.dart';
import 'package:universal_io/driver_base.dart';
import 'package:universal_io/io.dart';

class BrowserHttpClient extends BaseHttpClient {
  BrowserHttpClient(SecurityContext context);

  @override
  Future<HttpClientRequest> openUrl(String method, Uri url) async {
    if (url.host == null) {
      throw ArgumentError.value(url, "url", "Host can't be null");
    }
    var scheme = url.scheme;
    var needsNewUrl = false;
    if (scheme == null) {
      scheme = "https";
      needsNewUrl = true;
    } else {
      switch (scheme) {
        case "":
          scheme = "https";
          needsNewUrl = true;
          break;
        case "http":
          break;
        case "https":
          break;
        default:
          throw ArgumentError.value("Unsupported scheme '$scheme'");
      }
    }
    if (needsNewUrl) {
      url = Uri(
        scheme: scheme,
        userInfo: url.userInfo,
        host: url.host,
        port: url.port,
        query: url.query,
        fragment: url.fragment,
      );
    }
    final request = BrowserHttpClientRequest._(this, method, url);
    if (userAgent != null) {
      request.headers.add("User-Agent", userAgent);
    }
    return request;
  }

  /// Tells whether the request is cross-origin.
  static bool isCrossOriginUrl(String url, {String origin}) {
    origin ??= html.window.origin;

    // Add '/' so 'http://example.com' and 'http://example.com.other.com'
    // will be different.
    if (!origin.endsWith("/")) {
      origin = "$origin/";
    }

    return !url.startsWith(origin);
  }
}

class BrowserHttpClientException implements SocketException {
  static final Set<String> _corsSimpleMethods = Set<String>.from(const [
    "GET",
    "HEAD",
    "POST",
  ]);
  final String method;
  final String url;
  final String origin;

  final bool corsCredentialsMode;

  @override
  final String message = null;

  @override
  final OSError osError = null;

  @override
  final InternetAddress address = null;

  @override
  final int port = null;

  BrowserHttpClientException(
      {@required this.method,
      @required this.url,
      @required this.origin,
      @required this.corsCredentialsMode});

  @override
  String toString() {
    final sb = StringBuffer();
    for (var i = 0; i < 80; i++) {
      sb.write("-");
    }
    sb.write("\n");
    sb.write("""
BrowserHttpClient received an error from XMLHttpRequest (which doesn't tell
reason for the error).\n""");
    // Write a line
    sb.write("\n");

    // Write key details
    void addEntry(String key, String value) {
      sb.write(key.padRight(20));
      sb.write(value);
      sb.write("\n");
    }

    addEntry("HTTP method: ", method);
    addEntry("URL: ", url);
    addEntry("Origin: ", origin);

    // Warn about possible problem with missing CORS headers
    if (BrowserHttpClient.isCrossOriginUrl(url, origin: origin) ||
        corsCredentialsMode) {
      sb.write("\n");
      sb.write("Cross-origin request!\n");
      sb.write("'CORS credentials mode' is ");
      if (corsCredentialsMode) {
        sb.write("enabled (cookies will be supported).\n");
      } else {
        sb.write("disabled (cookies will NOT be supported).\n");
      }
      sb.write("\n");
      sb.write("""
If the URL is correct and the server actually responded, did the response
include the following required CORS headers?\n""");

      sb.write("  * Access-Control-Allow-Origin: $origin\n");
      if (corsCredentialsMode) {
        sb.write(
            "    * Wildcard '*' is not allowed because of credentials mode!\n");
        sb.write("  * Access-Control-Allow-Credentials: true\n");
      } else {
        sb.write("    * Wildcard '*' is also acceptable.\n");
      }
      if (!_corsSimpleMethods.contains(method)) {
        sb.write("  * Access-Control-Allow-Methods: $method\n");
      }
    }
    // Write a line
    for (var i = 0; i < 80; i++) {
      sb.write("-");
    }
    sb.write("\n");
    return sb.toString();
  }
}

class BrowserHttpClientRequest extends BaseHttpClientRequest {
  final HttpClient _client;

  @override
  final String method;

  @override
  final Uri uri;

  @override
  final HttpHeaders headers = HttpHeadersImpl("1.0");

  final Int8Buffer _buffer = Int8Buffer();

  final Completer<HttpClientResponse> _completer =
      Completer<HttpClientResponse>();
  bool _closed = false;
  Future _requestBodyFuture;

  @override
  final List<Cookie> cookies = <Cookie>[];

  bool useCorsCredentials = false;

  BrowserHttpClientRequest._(this._client, this.method, this.uri);

  @override
  HttpConnectionInfo get connectionInfo => null;

  @override
  Future<HttpClientResponse> get done {
    return _completer.future;
  }

  @override
  Encoding get encoding => utf8;

  @override
  set encoding(Encoding value) {
    throw StateError("IOSink encoding is not mutable");
  }

  @override
  void addError(Object error, [StackTrace stackTrace]) {
    if (_closed) {
      throw StateError("HTTP request is closed already");
    }
    _completer.completeError(error, stackTrace);
  }

  @override
  Future<void> addStream(Stream<List<int>> stream) {
    if (_requestBodyFuture != null) {
      throw StateError("StreamSink is bound to a stream");
    }
    if (_closed) {
      throw StateError("StreamSink is closed");
    }
    _requestBodyFuture = stream.listen((item) {
      _buffer.addAll(item);
    }, onError: (error) {
      addError(error);
    }, cancelOnError: true).asFuture(null);
    return _requestBodyFuture;
  }

  @override
  Future<void> flush() {
    return Future<void>.value();
  }

  @override
  void internallyAdd(List<int> event) {
    if (_requestBodyFuture != null) {
      throw StateError("StreamSink is bound to a stream");
    }
    if (_closed) {
      throw StateError("StreamSink is closed");
    }
    if (event.isEmpty) {
      return;
    }
    if (!_httpMethodSupportsBody(method)) {
      throw StateError("HTTP method $method does not support body");
    }
    _buffer.addAll(event);
  }

  @override
  Future<HttpClientResponse> internallyClose() {
    if (_closed) {
      throw StateError("StreamSink is closed");
    }
    if (!_completer.isCompleted) {
      if (_requestBodyFuture != null) {
        _requestBodyFuture.then((_) {
          return _send();
        });
      } else {
        _send();
      }
    }
    return _completer.future;
  }

  @override
  void write(Object obj) {
    String string = '$obj';
    if (string.isEmpty) return;
    add(utf8.encode(string));
  }

  @override
  void writeAll(Iterable objects, [String separator = ""]) {
    Iterator iterator = objects.iterator;
    if (!iterator.moveNext()) return;
    if (separator.isEmpty) {
      do {
        write(iterator.current);
      } while (iterator.moveNext());
    } else {
      write(iterator.current);
      while (iterator.moveNext()) {
        write(separator);
        write(iterator.current);
      }
    }
  }

  @override
  void writeCharCode(int charCode) {
    write(String.fromCharCode(charCode));
  }

  @override
  void writeln([Object object = ""]) {
    write(object);
    write("\n");
  }

  void _send() {
    try {
      final xhr = html.HttpRequest();

      // Set method and URI
      final uriString = uri.toString();
      xhr.open(method, uriString);

      // Set response body type
      xhr.responseType = 'blob';

      // Set timeout
      final timeout = this._client.connectionTimeout;
      if (timeout != null) {
        xhr.timeout = timeout.inMilliseconds;
      }

      // Do we have credentials enabled?
      // Or are we are trying to send cross-origin cookies?
      final origin = html.window.origin;
      var corsCredentialsMode = this.useCorsCredentials ||
          (BrowserHttpClient.isCrossOriginUrl(uriString, origin: origin) &&
              this.cookies.isNotEmpty);

      if (corsCredentialsMode) {
        // Yes
        xhr.withCredentials = true;
      }

      // Copy headers to html.HttpRequest
      headers.forEach((name, values) {
        for (var value in values) {
          xhr.setRequestHeader(name, value);
        }
      });

      final completer = this._completer;

      xhr.onReadyStateChange.listen((event) {
        if (xhr.readyState == html.HttpRequest.LOADING) {}
      });

      // Set response handler
      xhr.onLoadEnd.first.then((_) async {
        // Read body
        final body = await _readResponseBody(method, uri, xhr);
        if (completer.isCompleted) {
          // An error occurred during reading
          return;
        }

        // Create HttpClientResponse
        final httpClientResponse = BrowserHttpClientResponse._(
          _client,
          this,
          method,
          uri,
          headers,
          xhr,
          body,
        );

        // Complete the future
        completer.complete(httpClientResponse);
      });

      xhr.onError.listen((html.ProgressEvent event) {
        if (completer.isCompleted) {
          return;
        }

        // Unfortunately, the underlying XMLHttpRequest API doesn't expose any
        // specific information about the error itself.
        //
        // We gather the information that we have and try to produce a
        // descriptive exception.
        final error = BrowserHttpClientException(
          method: method,
          url: uriString,
          origin: origin,
          corsCredentialsMode: corsCredentialsMode,
        );

        completer.completeError(error, StackTrace.current);
      });

      if (_httpMethodSupportsBody(method)) {
        // Send with body
        xhr.send(Uint8List.fromList(_buffer));
      } else {
        // Send without body
        xhr.send();
      }
    } catch (e) {
      // Something went wrong
      _completer.completeError(e);
    }
  }

  static bool _httpMethodSupportsBody(String method) {
    switch (method) {
      case "GET":
        return false;
      case "HEAD":
        return false;
      case "OPTIONS":
        return false;
      default:
        return true;
    }
  }

  static Stream<List<int>> _readResponseBody(
      String method, Uri uri, html.HttpRequest request) {
    final blob = request.response;
    if (blob == null) {
      // No body
      return Stream<List<int>>.empty();
    }

    // Read response with FileReader
    final controller = StreamController<Uint8List>();
    var fileReader = html.FileReader();
    fileReader.onLoad.first.then((html.ProgressEvent event) {
      var body = fileReader.result as Uint8List;
      controller.add(body);
      controller.close();
    });

    fileReader.onError.listen((html.ProgressEvent event) {
      controller.addError(
        SocketException("FileReader error. Request was: '$method $uri'"),
        StackTrace.current,
      );
      controller.close();
    });
    fileReader.readAsArrayBuffer(blob);
    return controller.stream;
  }
}

class BrowserHttpClientResponse extends BaseHttpClientResponse {
  final HttpClient client;
  final HttpClientRequest request;
  final String _requestMethod;
  final HttpHeaders _requestHeaders;
  final html.HttpRequest _response;
  final Stream<List<int>> _body;

  @override
  final HttpHeaders headers = HttpHeadersImpl("1.0");

  BrowserHttpClientResponse._(this.client, this.request, this._requestMethod,
      Uri requestUri, this._requestHeaders, this._response, this._body)
      : assert(_response != null) {
    final headers = this.headers;
    _response.responseHeaders.forEach((k, v) {
      headers.add(k, v);
    });
  }

  @override
  bool get isRedirect =>
      HttpStatus.temporaryRedirect == statusCode ||
      HttpStatus.movedPermanently == statusCode;

  @override
  String get reasonPhrase {
    return _response.statusText;
  }

  int get statusCode => _response.status;

  @override
  StreamSubscription<List<int>> listen(void onData(List<int> event),
      {Function onError, void onDone(), bool cancelOnError}) {
    return _body.listen(
      onData,
      onError: onError,
      onDone: onDone,
      cancelOnError: cancelOnError,
    );
  }

  @override
  Future<HttpClientResponse> redirect(
      [String method, Uri url, bool followLoops]) {
    final newUrl =
        url ?? Uri.parse(this.headers.value(HttpHeaders.locationHeader));
    return client.openUrl(method ?? _requestMethod, newUrl).then((newRequest) {
      _requestHeaders.forEach((name, value) {
        newRequest.headers.add(name, value);
      });
      newRequest.followRedirects = true;
      return newRequest.close();
    });
  }
}
